#!/bin/bash

BUF=$QUERY_STRING\&

SKETCH_PARAMS=
BRIM_PARAMS=

while [[ -n $BUF ]]; do
{
  KEY=`echo $BUF | awk '{ print substr($0,0,match($0,"=")); }'`
  BUF=`echo $BUF | awk '{ print substr($0,match($0,"=")+1); }'`
  VALUE=`echo $BUF | awk '{ print substr($0,0,match($0,"\&")); }'`
  BUF=`echo $BUF | awk '{ print substr($0,match($0,"\&")+1); }'`
  if [[ $KEY == "network" && -n $VALUE ]]; then
  {
    NETWORK=$VALUE
    NETWORK_=`echo $VALUE | ../bin/uncgi`
  };
  elif [[ $KEY == "ref" && -n $VALUE ]]; then
  {
    REF=$VALUE
    REF_=`echo $VALUE | ../bin/uncgi`
  };
  elif [[ $KEY == "width" && -n $VALUE ]]; then
    SKETCH_PARAMS="$SKETCH_PARAMS --width=$VALUE"
  elif [[ $KEY == "height" && -n $VALUE ]]; then
    SKETCH_PARAMS="$SKETCH_PARAMS --height=$VALUE"
  elif [[ $KEY == "font-size" && -n $VALUE ]]; then
    SKETCH_PARAMS="$SKETCH_PARAMS --stop-font-size=$VALUE"
  elif [[ $KEY == "force-rows" && -n $VALUE ]]; then
    SKETCH_PARAMS="$SKETCH_PARAMS --rows=$VALUE"
  elif [[ $KEY == "style" && -n $VALUE ]]; then
  {
    SKETCH_PARAMS="$SKETCH_PARAMS --options=/opt/osm_why_api/options/sketch-line.$VALUE"
    BRIM_PARAMS="$BRIM_PARAMS --options=/opt/osm_why_api/options/sketch-line.$VALUE"
  };
  elif [[ $KEY == "correspondences" && -n $VALUE ]]; then
  {
    SKETCH_PARAMS="$SKETCH_PARAMS --walk-limit=$VALUE"
    BRIM_PARAMS="$BRIM_PARAMS --size=$VALUE"
  };
  elif [[ $KEY == "max-cors-per-line" && -n $VALUE ]]; then
    SKETCH_PARAMS="$SKETCH_PARAMS --max-correspondences-per-line=$VALUE"
  elif [[ $KEY == "max-cors-below" && -n $VALUE ]]; then
    SKETCH_PARAMS="$SKETCH_PARAMS --max-correspondences-below=$VALUE"
  elif [[ $KEY == "debug" && -n $VALUE ]]; then
    DEBUG=$VALUE
  fi
};
done

BASEDIR=`mktemp -d`

echo -e "\
data=<osm-script> \
 \
<union> \
  <query type=\"relation\"> \
    <has-kv k=\"network\" v=\"$NETWORK_\"/> \
    <has-kv k=\"ref\" v=\"$REF_\"/> \
  </query> \
  <recurse type=\"relation-node\"/> \
</union> \
<print mode=\"body\"/> \
 \
</osm-script>
" >$BASEDIR/request.1

if [[ -z $REF ]]; then
{                     
  echo "Content-Type: text/plain; charset=utf-8"
  echo                                          
  echo "An empty value for ref is not allowed"  

  exit 0
};    
fi      

CORRESPONDENCES=`../bin/bbox-brim-query --only-corrs $BRIM_PARAMS`

if [[ $CORRESPONDENCES -gt 0 ]]; then
{
  REQUEST_METHOD=
  /home/roland/osm-3s/build/bin/osm3s_query --quiet --no-mime <$BASEDIR/request.1 >$BASEDIR/answer.1
  ../bin/bbox-brim-query $BRIM_PARAMS <$BASEDIR/answer.1 >$BASEDIR/request.2

  if [[ $DEBUG == "full-query" ]]; then
  {
    echo "Content-Type: text/plain; charset=utf-8"
    echo
    cat <$BASEDIR/request.2
    echo
  };
  fi

  REQUEST_METHOD=
  /home/roland/osm-3s/build/bin/osm3s_query --quiet --no-mime <$BASEDIR/request.2 >$BASEDIR/answer.3

  echo "Content-Type: image/svg+xml; charset=utf-8"
  echo

  if [[ $DEBUG == "full-query" ]]; then
  {
    cat <$BASEDIR/answer.3
    echo
    echo "../bin/sketch-route-svg --ref=\"$REF_\" --network=\"$NETWORK_\" $SKETCH_PARAMS"
    exit 0;
  };
  fi

  ../bin/sketch-route-svg --ref="$REF_" --network="$NETWORK_" $SKETCH_PARAMS <$BASEDIR/answer.3
};
else
{
  REQUEST_METHOD=
  /home/roland/osm-3s/build/bin/osm3s_query --quiet --no-mime <$BASEDIR/request.1 >$BASEDIR/answer.1

  if [[ $DEBUG == "full-query" ]]; then
  {
    echo "Content-Type: text/plain; charset=utf-8"
    echo
    cat <$BASEDIR/answer.1
    echo
  };
  fi;

  echo "Content-Type: image/svg+xml; charset=utf-8"
  echo

  ../bin/sketch-route-svg $SKETCH_PARAMS <$BASEDIR/answer.1
};
fi
